# :: AUTOGEN-START ::

from .arc import ArcType, ArcColor, Arctap, Arc, ScaledArctap
from .camera import CameraEasing, Camera
from .hold import Hold, HoldFloat
from .scenecontrol import SceneControlType, SceneControl
from .tap import Tap, TapFloat
from .timing import Timing

__all__ = [
    'ArcType', 
    'ArcColor', 
    'Arctap', 
    'Arc', 
    'ScaledArctap', 
    'CameraEasing', 
    'Camera', 
    'Hold', 
    'HoldFloat', 
    'SceneControlType', 
    'SceneControl', 
    'Tap', 
    'TapFloat', 
    'Timing', 
]

# :: AUTOGEN-END ::
